/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Vector.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhBinarize;
   
DECLARE_NODE(CvfdhBinarize)
/*Node
 *
 * @name CvfdhBinarize
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhBinarize
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name THRESH
 * @input_type int
 * @input_description Threshold setting for binarization.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 * 
 * @parameter_name PARAM
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Threshold setting for binalization. This setting will be cancelled when the input THRESH exists.
 * 
 * @parameter_name BW_SELECT
 * @parameter_type string
 * @parameter_list BLACK:WHITE
 * @parameter_value WHITE
 * @parameter_description Black-white selection if the pixel value is more than threshold. 
 *
 END*/
   
   
class CvfdhBinarize : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  int threshID;
  int thresh;
  string bw_select;
      
public:
  CvfdhBinarize(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), threshID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    thresh = dereference_cast<int>(parameters.get("PARAM"));
    bw_select = object_cast<String>(parameters.get("BW_SELECT"));
    inOrder = true;
  } 

  virtual int translateInput(string inputName) {
    if (inputName == "THRESH") {
      threshID = addInput(inputName);
      return threshID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    ObjectRef in = getInput(m_imageInID, count);

    if(threshID != -1){
      ObjectRef input = getInput(threshID, count);
      thresh = dereference_cast<int> (input);
    }
    
    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }
      IplImage* img_in;
      img_in = cvCloneImage(imagePtr->getImage());
      IplImage *dst_img = NULL;
      dst_img = cvCloneImage (img_in);

      int nch = img_in->nChannels;
      for(int ch = 1; ch <= nch; ch++){
	cvSetImageCOI( dst_img, ch );
	IplImage *cp_img = NULL;
	cp_img  = cvCreateImage( cvGetSize(dst_img),dst_img->depth,1);
	cvCopy(dst_img, cp_img);	  
	MyBinarize(cp_img, thresh, bw_select);
	cvCopy(cp_img, dst_img);
	cvReleaseImage (&cp_img);	  
      }
      cvSetImageCOI( dst_img, 0 );

      IplImage* img_out;
      img_out = cvCloneImage(dst_img);
      CvfdImage* image= new CvfdImage(img_out);     
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image);

      cvReleaseImage (&img_in);
      cvReleaseImage (&dst_img);
      cvReleaseImage (&img_out);

      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* img_in;
	img_in = cvCloneImage(imagePtr.getImage());
	IplImage *dst_img = 0;
	dst_img = cvCloneImage (img_in);

	int nch = img_in->nChannels;
	for(int ch = 1; ch <= nch; ch++){
	  cvSetImageCOI( dst_img, ch );
	  IplImage *cp_img = NULL;
	  cp_img  = cvCreateImage( cvGetSize(dst_img),dst_img->depth,1);
	  cvCopy(dst_img, cp_img);	  
	  MyBinarize(cp_img, thresh, bw_select);
	  cvCopy(cp_img, dst_img);
	  cvReleaseImage (&cp_img);	  
	}
	cvSetImageCOI( dst_img, 0 );
	
	IplImage* img_out;
	img_out = cvCloneImage(dst_img);
	CvfdImage* image= new CvfdImage(img_out);  
	(*outputMapPtr)[it->first] = image;
	cvReleaseImage (&img_in);
	cvReleaseImage (&dst_img);
	cvReleaseImage (&img_out);
      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
            
    }
    
  }

  void MyBinarize(IplImage* pImg, int thresh, string selector)
  {
    if(pImg->depth == IPL_DEPTH_8U){
      for(int cnt=0;cnt<pImg->width*pImg->height;cnt++){
	unsigned char* ptr = (unsigned char*)pImg->imageData+cnt;
	if(selector == "WHITE") *ptr = (*ptr >= (unsigned char)thresh)?0xff:0x00;
	if(selector == "BLACK") *ptr = (*ptr >= (unsigned char)thresh)?0x00:0xff;
      }
    } else if (pImg->depth == IPL_DEPTH_16U) {
      for(int cnt=0;cnt<pImg->width*pImg->height;cnt++){
	unsigned short* ptr = (unsigned short*)pImg->imageData+cnt;
	if(selector == "WHITE") *ptr = (*ptr >= (unsigned short)thresh)?0xffff:0x0000;
	if(selector == "BLACK") *ptr = (*ptr >= (unsigned short)thresh)?0x0000:0xffff;
      }      
    }
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhBinarize)
};
