/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdContours.h"
#include "CvfdImage.h"
#include "CvfdColor.h"

using namespace std;
using namespace FD;

class CvfdGravityCenter;
   
DECLARE_NODE(CvfdGravityCenter)
/*Node
 *
 * @name CvfdGravityCenter
 * @category FDOPENCV_ORG:Arithmetic Logic Comparison
 * @description Calculate the coordinate of the gravity center.
 *
 * @input_name IMAGE
 * @input_description The image to handle
 * @input_type CvfdImage
 *
 * @parameter_name BINARY
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description If the flag is non-zero, all the zero pixel values are treated as zeroes, all the others are treated as 0xFF.
 *
 * @parameter_name CHANNEL
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description Sets the channel of interest to a given value. Value 0 means that all channels are selected, 1 means that the first channel is selected etc.
 *
 * @output_name X
 * @output_description The coordinate x of the gravity center.
 * @output_type float
 *
 * @output_name Y
 * @output_description The coordinate y of the gravity center.
 * @output_type float
 *
 END*/
   
class CvfdGravityCenter : public BufferedNode {
      
  //Input ID
  int m_imageID;
  //Output ID
  int m_xID;
  int m_yID;
  //Parameters
  int m_binary;
  int m_channel;
      
public:
  CvfdGravityCenter(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_imageID = addInput("IMAGE");
         
    //add outputs
    m_xID = addOutput("X");
    m_yID = addOutput("Y");
         
    //Initialize parameters         
    m_binary = dereference_cast<int>(parameters.get("BINARY"));
    m_channel = dereference_cast<int>(parameters.get("CHANNEL"));
         
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {           
    //Read the inputs
    RCPtr<CvfdImage> imagePtr = getInput(m_imageID,count);  
         
    //Handle
    CvMoments* moments = new CvMoments();
    CvfdImage* image = new CvfdImage(imagePtr->getImage());
    float x=0,y=0;
         
    int status = cvGetErrMode();
         
    cvSetErrMode( CV_ErrModeSilent );
    __BEGIN__;          
    OPENCV_CALL( cvSetImageCOI( image->getImage(), m_channel) );
    OPENCV_CALL( cvMoments( image->getImage(), moments, m_binary)); 
    if(cvGetSpatialMoment( moments, 0, 0 )!=0){
      OPENCV_CALL( x = (float)(cvGetSpatialMoment( moments, 1, 0 )/cvGetSpatialMoment( moments, 0, 0 )));
      OPENCV_CALL( y = (float)(cvGetSpatialMoment( moments, 0, 1 )/cvGetSpatialMoment( moments, 0, 0 )));
    }
    __END__;
    cvSetErrMode( status );
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to calculate the gravity center: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      } 
    delete image;
    (*(outputs[m_xID].buffer))[count] = ObjectRef(Float::alloc(x)); 
    (*(outputs[m_yID].buffer))[count] = ObjectRef(Float::alloc(y)); 
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdGravityCenter)
};
